"""
Code relating to the command line interface to acclimatise
"""
import argparse
import subprocess
import sys
import typing

from pyparsing import ParseBaseException

from acclimatise import best_cmd, parse_help
from acclimatise.converter import WrapperGenerator, cases
from acclimatise.model import Command


def main():
    args = get_parser().parse_args()

    # Allow input of help text either by running the command, or using stdin
    if args.stdin:
        stdin = "".join(sys.stdin.readlines())
        command = parse_help(args.cmd, stdin)
    else:
        command = best_cmd(args.cmd)

    converter_cls = WrapperGenerator.choose_converter(args.format)
    converter = converter_cls(
        generate_names=args.generate_names,
        ignore_positionals=args.no_pos,
        case=args.case,
    )
    output = converter.generate_wrapper(command)
    print(output)


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("cmd", nargs="+", help="The base command this tool uses")
    parser.add_argument(
        "-f",
        "--format",
        choices=["wdl", "cwl"],
        required=True,
        help="The language in which to output the CLI wrapper",
    )
    parser.add_argument(
        "--no-pos",
        action="store_true",
        help=(
            "Don't include positional arguments, for example because the help formatting has some "
            "misleading sections that look like positional arguments"
        ),
    )
    parser.add_argument(
        "-g",
        "--generate-names",
        action="store_true",
        help=(
            "Rather than using the long flag to generate the argument name, generate them automatically using the "
            "flag description. Generally helpful if there are no long flags, only short flags."
        ),
    )
    parser.add_argument(
        "-c",
        "--case",
        choices=cases,
        help=(
            "Which case to use for variable names. If not set, defaults to the language defaults: snake_case for CWL"
            " and snake_case for WDL"
        ),
        default="snake",
    )
    parser.add_argument(
        "--stdin",
        help="Accept the help text from stdin instead of running the command",
        action="store_true",
    )
    return parser


if __name__ == "__main__":
    main()



"""
Functions for generating WDL from the CLI data model
"""
import re
import subprocess
from dataclasses import dataclass

from acclimatise import cli_types, model
from acclimatise.converter import WrapperGenerator
from acclimatise.model import Command
from inflection import camelize
from wdlgen import ArrayType, Input, PrimitiveType, Task, WdlType


def flag_to_command_input(flag: model.CliArgument) -> Task.Command.CommandInput:
    args = dict(name=flag.name_to_camel())

    if isinstance(flag, model.Flag):
        args.update(dict(optional=True,))
        if isinstance(flag.args, model.EmptyFlagArg):
            args.update(dict(true=flag.longest_synonym, false=""))
        else:
            args.update(dict(prefix=flag.longest_synonym,))
    elif isinstance(flag, model.Positional):
        args.update(dict(optional=False, position=flag.position))

    return Task.Command.CommandInput(**args)


class WdlGenerator(WrapperGenerator):
    case = "snake"

    @classmethod
    def format(cls) -> str:
        return "wdl"

    @classmethod
    def type_to_wdl(cls, typ: cli_types.CliType, optional: bool = False) -> WdlType:
        if isinstance(typ, cli_types.CliString):
            return WdlType(PrimitiveType(PrimitiveType.kString), optional=optional)
        elif isinstance(typ, cli_types.CliFloat):
            return WdlType(PrimitiveType(PrimitiveType.kFloat), optional=optional)
        elif isinstance(typ, cli_types.CliBoolean):
            return WdlType(PrimitiveType(PrimitiveType.kBoolean), optional=optional)
        elif isinstance(typ, cli_types.CliInteger):
            return WdlType(PrimitiveType(PrimitiveType.kInt), optional=optional)
        elif isinstance(typ, cli_types.CliFile):
            return WdlType(PrimitiveType(PrimitiveType.kFile), optional=optional)
        elif isinstance(typ, cli_types.CliDir):
            return WdlType(PrimitiveType(PrimitiveType.kDirectory), optional=optional)
        elif isinstance(typ, cli_types.CliTuple):
            if typ.homogenous:
                return WdlType(
                    ArrayType(
                        cls.type_to_wdl(typ.values[0]), requires_multiple=not optional
                    )
                )
            else:
                raise Exception("WDL can't deal with hetrogenous array types")
        elif isinstance(typ, cli_types.CliList):
            return WdlType(
                ArrayType(cls.type_to_wdl(typ.value), requires_multiple=not optional)
            )
        elif isinstance(typ, cli_types.CliEnum):
            return WdlType(PrimitiveType(PrimitiveType.kString), optional=optional)
        else:
            return WdlType(PrimitiveType(PrimitiveType.kString), optional=optional)

    def formulate_command(self, cmd: Command) -> str:
        args = subprocess.list2cmdline(cmd.command)

        for flag in cmd.named:
            args += " \\\n\t"
            if isinstance(flag.args.get_type(), cli_types.CliBoolean):
                args += '~{{true="{}" false="" {}}}'.format(
                    flag.longest_synonym, self.choose_variable_name(flag)
                )
            else:
                args += '~{{"{} " + {}}}'.format(
                    flag.longest_synonym, self.choose_variable_name(flag)
                )

        return args

    @staticmethod
    def sanitize_wdl_str(string: str):
        return re.sub(
            "\)",
            "]",  # Replace left paren with a bracket
            re.sub(
                "\(",
                "[",  # Replace right paren with a bracket
                re.sub(
                    '"',
                    "'",  # Remote double quote string
                    re.sub("[{}]", "", string),  # Remove braces
                ),
            ),
        )

    def generate_wrapper(self, cmd: Command) -> str:
        name = camelize("_".join(cmd.command).replace("-", "_"))

        inputs = [
            Input(
                data_type=self.type_to_wdl(pos.get_type(), optional=False),
                name=pos.name_to_camel(),
            )
            for pos in cmd.named
        ]
        if not self.ignore_positionals:
            inputs += [
                Input(
                    data_type=self.type_to_wdl(pos.get_type(), optional=True),
                    name=pos.name_to_camel(),
                )
                for pos in cmd.positional
            ]

        tool = Task(
            name=name,
            command=Task.Command(
                command=" ".join(cmd.command),
                inputs=[flag_to_command_input(pos) for pos in cmd.positional],
                arguments=[flag_to_command_input(named) for named in cmd.named],
            ),
            version="1.0",
            inputs=inputs,
        )

        return tool.get_string()

        # return template.render(
        #     taskname=name,
        #     positional=cmd.positional,
        #     named=cmd.named,
        #     command=self.formulate_command(cmd),
        #     generate_names=self.generate_names
        # )



from abc import abstractmethod
from dataclasses import dataclass

from acclimatise.model import CliArgument, Command

cases = ["snake", "camel"]


@dataclass
class WrapperGenerator:
    """
    Abstract base class for a class that converts a Command object into a string that defines a tool
    wrapper in a certain workflow language
    """

    @classmethod
    def choose_converter(cls, typ):
        """
        Returns a converter subclass, given a converter type name
        :param type: The type of converter, e.g. 'cwl' or 'wdl'
        """
        for subclass in cls.__subclasses__():
            if subclass.format() == typ:
                return subclass

        raise Exception("Unknown format type")

    @classmethod
    @abstractmethod
    def format(cls) -> str:
        """
        Returns the output format that this generator produces as a string, e.g. "cwl"
        """
        pass

    @abstractmethod
    def generate_wrapper(self, cmd: Command) -> str:
        """
        Abstract method that defines the interface for converting a Command into a wrapper string
        """
        pass

    def choose_variable_name(self, flag: CliArgument) -> str:
        """
        Choose a name for this flag (e.g. the variable name when this is used to generate code), based on whether
        the user wants an auto generated one or not
        """
        if self.generate_names:
            return flag.generate_name()
        elif self.case == "snake":
            return flag.name_to_snake()
        elif self.case == "camel":
            return flag.name_to_camel()

    case: str = "snake"
    """
    Which case to use for variable names
    """

    generate_names: bool = False
    """
    Rather than using the long flag to generate the argument name, generate them automatically using the
    flag description. Generally helpful if there are no long flags, only short flags.
    """

    ignore_positionals: bool = False
    """
    Don't include positional arguments, for example because the help formatting has some
    misleading sections that look like positional arguments
    """



import tempfile
from dataclasses import dataclass

import cwlgen
from acclimatise import cli_types
from acclimatise.converter import WrapperGenerator
from acclimatise.model import Command


@dataclass
class CwlGenerator(WrapperGenerator):
    case = "snake"

    @classmethod
    def format(cls) -> str:
        return "cwl"

    @staticmethod
    def snake_case(words: list):
        return "_".join([word.lower() for word in words])

    @staticmethod
    def to_cwl_type(typ: cli_types.CliType):
        if isinstance(typ, cli_types.CliFile):
            return "File"
        elif isinstance(typ, cli_types.CliDir):
            return "Directory"
        elif isinstance(typ, cli_types.CliString):
            return "string"
        elif isinstance(typ, cli_types.CliFloat):
            return "double"
        elif isinstance(typ, cli_types.CliInteger):
            return "long"
        elif isinstance(typ, cli_types.CliBoolean):
            return "boolean"
        elif isinstance(typ, cli_types.CliEnum):
            return "string"
        elif isinstance(typ, cli_types.CliList):
            return CwlGenerator.to_cwl_type(typ.value) + "[]"
        elif isinstance(typ, cli_types.CliTuple):
            return [CwlGenerator.to_cwl_type(subtype) for subtype in set(typ.values)]
        else:
            raise Exception(f"Invalid type {typ}!")

    def generate_wrapper(self, cmd: Command) -> str:
        cwl_tool = cwlgen.CommandLineTool(
            tool_id=self.snake_case(cmd.command),
            base_command=cmd.command,
            cwl_version="v1.0",
        )

        if not self.ignore_positionals:
            for pos in cmd.positional:
                cwl_tool.inputs.append(
                    cwlgen.CommandInputParameter(
                        param_id=self.choose_variable_name(pos),
                        param_type=self.to_cwl_type(pos.get_type()),
                        input_binding=cwlgen.CommandLineBinding(position=pos.position),
                        doc=pos.description,
                    )
                )

        for flag in cmd.named:
            cwl_tool.inputs.append(
                cwlgen.CommandInputParameter(
                    param_id=self.choose_variable_name(flag),
                    param_type=self.to_cwl_type(flag.get_type()),
                    input_binding=cwlgen.CommandLineBinding(
                        prefix=flag.longest_synonym
                    ),
                    doc=flag.description,
                )
            )

        with tempfile.NamedTemporaryFile(mode="w+", encoding="utf-8") as temp:
            cwl_tool.export(temp.name)
            return temp.read()



from io import StringIO

from acclimatise import parse_help
from acclimatise.yaml import yaml


def test_round_trip(bwamem_help):
    command = parse_help(["bwa", "mem"], bwamem_help)

    # Dump
    buffer = StringIO()
    yaml.dump(command, buffer)

    # Load
    buffer.seek(0)
    output = yaml.load(buffer)

    # Assert the round trip worked
    assert command == output



"""
Re-usable parser elements that aren't tied to the parser object
"""
from pyparsing import *
from pyparsing import _bslash

from acclimatise.model import *


def customIndentedBlock(
    blockStatementExpr, indentStack, indent=True, terminal=False, lax=False
):
    """
    Modified version of the indentedBlock construct provided by pyparsing. Allows fuzzier indent boundaries
    """
    backup_stack = indentStack[:]

    def reset_stack():
        indentStack[:] = backup_stack

    def checkPeerIndent(s, l, t):
        if l >= len(s):
            return
        curCol = col(l, s)

        # A terminal block doesn't have children, so we can assume that any sub-indent is a peer
        if terminal and curCol >= indentStack[-1]:
            return

        # If we're being lax, anything that's not a full dedent is a peer
        if lax and curCol > indentStack[-2]:
            return

        # Anything that is indented more than the previous indent level counts as a peer
        if curCol < indentStack[-1] or curCol <= indentStack[-2]:
            raise ParseException(s, l, "not a peer entry")

    def checkSubIndent(s, l, t):
        curCol = col(l, s)
        if curCol > indentStack[-1]:
            indentStack.append(curCol)
        else:
            raise ParseException(s, l, "not a subentry")

    def checkUnindent(s, l, t):
        if l >= len(s):
            indentStack.pop()
            return
        curCol = col(l, s)
        if not (indentStack and curCol < indentStack[-1]):
            raise ParseException(s, l, "not an unindent")
        indentStack.pop()

    NL = OneOrMore(LineEnd().setWhitespaceChars("\t ").suppress())
    INDENT = (Empty() + Empty().setParseAction(checkSubIndent)).setName("INDENT")
    PEER = Empty().setParseAction(checkPeerIndent).setName("")
    UNDENT = Empty().setParseAction(checkUnindent).setName("UNINDENT")
    if indent:
        smExpr = Group(
            Optional(NL)
            +
            # ~ FollowedBy(blockStatementExpr) +
            INDENT
            + (OneOrMore(PEER + Group(blockStatementExpr) + Optional(NL)))
            + UNDENT
        )
    else:
        smExpr = Group(
            Optional(NL) + (OneOrMore(PEER + Group(blockStatementExpr) + Optional(NL)))
        )
    smExpr.setFailAction(lambda a, b, c, d: reset_stack())
    blockStatementExpr.ignore(_bslash + LineEnd())
    return smExpr.setName("custom indented block")


cli_id = Word(initChars=alphas + "@", bodyChars=alphanums + "-_@")

# short_flag = originalTextFor(Literal('-') + Word(alphanums + '@', max=1))
# """A short flag has only a single dash and single character, e.g. `-m`"""
# long_flag = originalTextFor(Literal('--') + cli_id)
# """A long flag has two dashes and any amount of characters, e.g. `--max-count`"""
any_flag = originalTextFor("-" + Optional("-") + cli_id)
"""The flag is the part with the dashes, e.g. `-m` or `--max-count`"""

flag_arg_sep = Or([Literal("="), Literal(" ")]).leaveWhitespace()
"""The term that separates the flag from the arguments, e.g. in `--file=FILE` it's `=`"""

arg = cli_id.copy()
"""A single argument name, e.g. `FILE`"""

optional_args = Forward()


def visit_optional_args(s, lok, toks):
    if len(toks) == 1:
        return OptionalFlagArg(names=[toks[0]])
    else:
        other = toks[3]
        if isinstance(other, str):
            return OptionalFlagArg(names=[toks[0], other])
        elif isinstance(other, OptionalFlagArg):
            return OptionalFlagArg(names=[toks[0]] + other.names)


optional_args <<= (arg + "[" + "," + (optional_args ^ arg) + "]").setParseAction(
    visit_optional_args
)
"""
When the flag has multiple arguments, some of which are optional, e.g.
-I FLOAT[,FLOAT[,INT[,INT]]]
"""

# simple_arg = arg.copy().setParseAction(
#     lambda s, loc, toks: SimpleFlagArg(toks[0]))
simple_arg = Or(
    [
        Word(initChars=alphas, bodyChars=alphanums + "-_."),
        # Allow spaces in the argument name, but only if it's enclosed in angle brackets
        Literal("<").suppress()
        + Word(initChars=alphas, bodyChars=alphanums + "-_. ")
        + Literal(">").suppress(),
    ]
).setParseAction(lambda s, loc, toks: SimpleFlagArg(toks[0]))

list_type_arg = (
    arg
    + Literal(" ")
    + Literal("[")
    + matchPreviousLiteral(arg)
    + Optional(Literal(" "))
    + Literal("...")
    + Literal("]")
).setParseAction(lambda s, loc, toks: RepeatFlagArg(toks[0]))
"""When the argument is an array of values, e.g. when the help says `--samout SAMOUTS [SAMOUTS ...]`"""

choice_type_arg = nestedExpr(
    opener="{", closer="}", content=delimitedList(cli_id, delim=",")
).setParseAction(lambda s, loc, toks: ChoiceFlagArg(list(toks[0])))
"""When the argument is one from a list of values, e.g. when the help says `--format {sam,bam}`"""


def noop(s, loc, toks):
    return toks


arg_expression = (
    (
        flag_arg_sep.suppress()
        + (list_type_arg ^ choice_type_arg ^ optional_args ^ simple_arg)
    )
    .leaveWhitespace()
    .setParseAction(lambda s, loc, toks: toks[0])
)
"""An argument with separator, e.g. `=FILE`"""

flag_with_arg = (any_flag + Optional(arg_expression)).setParseAction(
    lambda s, loc, toks: (
        FlagSynonym(name=toks[0], argtype=toks[1] if len(toks) > 1 else EmptyFlagArg())
    )
)
"""e.g. `--max-count=NUM`"""

# TODO: this should be smarter, accepting ' ' or '| ' or '|' etc
synonym_delim = Word(" ,|", max=2).setParseAction(noop).leaveWhitespace()
"""
The character used to separate synonyms of a flag. Depending on the help text this might be a comma, pipe or space
"""

description_sep = White(min=1).setName("description_sep").suppress()
"""
The section that separates a flag from its description. This needs to be broad enough that it will match all different
formats of help outputs but not so broad that every single word starting with a dash will be matched as a flag
"""

# block_element_prefix = LineStart().leaveWhitespace()
block_element_prefix = (
    ((LineStart().leaveWhitespace() ^ Literal(":")) + White(min=1))
    .setName("block_element_prefix")
    .leaveWhitespace()
    .suppress()
)
"""
Each element (e.g. flag) in a list of flags must either start with a colon or nothing

e.g. in this example "index" is prefixed by a colon and "mem" is prefixed by a LineStart

Command: index         index sequences in the FASTA format
         mem           BWA-MEM algorithm
"""

flag_synonyms = delimitedList(flag_with_arg, delim=synonym_delim)
"""
When the help lists multiple synonyms for a flag, e.g:
-n, --lines=NUM
"""


# The description of the flag
# e.g. for grep's `-o, --only-matching`, this is:
# "Print only the matched (non-empty) parts of a matching line, with each such part on a separate output line."
def success(a, b, c):
    pass


desc_line = originalTextFor(SkipTo(LineEnd()))  # .setParseAction(success))



from textwrap import dedent

from pkg_resources import resource_filename

import pytest
from acclimatise.flag_parser.parser import CliParser


@pytest.fixture
def htseq_help():
    with open(resource_filename(__name__, "test_data/htseq_count.txt")) as fp:
        return fp.read()


@pytest.fixture
def bwamem_help():
    with open(resource_filename(__name__, "test_data/bwa_mem.txt")) as fp:
        return fp.read()


@pytest.fixture
def pisces_help():
    with open(resource_filename(__name__, "test_data/pisces.txt")) as fp:
        return fp.read()


@pytest.fixture
def bwa_help():
    with open(resource_filename(__name__, "test_data/bwa.txt")) as fp:
        return fp.read()


@pytest.fixture
def bedtools_help():
    with open(resource_filename(__name__, "test_data/bedtools.txt")) as fp:
        return fp.read()


@pytest.fixture
def bedtools_coverage_help():
    with open(resource_filename(__name__, "test_data/bedtools_coverage.txt")) as fp:
        return fp.read()


@pytest.fixture
def podchecker_help():
    with open(resource_filename(__name__, "test_data/podchecker.txt")) as fp:
        return fp.read()


@pytest.fixture()
def process():
    def process_help_section(help):
        """
        Does some preprocessing on a help text segment to facilitate testing
        """
        help = help.strip("\n")
        return dedent(help)

    return process_help_section



#!/usr/bin/env python3
from setuptools import find_packages, setup

setup(
    packages=find_packages(exclude="test"),
    install_requires=[
        "pyparsing",
        "jinja2",
        "spacy",
        "cwlgen",
        "dataclasses",
        "miniwdl",
        "wordsegment",
        "inflection",
        "illusional.wdlgen~=0.2",
        "ruamel.yaml>=0.16.5",
    ],
    entry_points={"console_scripts": ["acclimatise = acclimatise.cli:main"]},
    extras_require={"dev": ["pytest", "pre-commit"],},
)



import dataclasses
import subprocess
import typing

from pyparsing import ParseBaseException

from acclimatise.converter.cwl import CwlGenerator
from acclimatise.converter.wdl import WdlGenerator
from acclimatise.flag_parser.parser import CliParser
from acclimatise.model import Command
from acclimatise.usage_parser import parse_usage


def parse_help(cmd: typing.Collection[str], text: str, parse_positionals=True):
    help_command = CliParser(parse_positionals=parse_positionals).parse_command(
        name=cmd, cmd=text
    )
    usage_command = parse_usage(cmd, text)

    # Combine the two commands by picking from the help_command where possible, otherwise falling back on the usage
    fields = {}
    for field in dataclasses.fields(Command):
        fields[field.name] = getattr(help_command, field.name) or getattr(
            usage_command, field.name
        )
    command = Command(**fields)

    # Normally parsing the list of flags will provide a better command summary, but if it didn't give us anything,
    # fall back to the usage
    # if len(help_command.positional) + len(help_command.named) == 0:
    #     command = usage_command
    # else:
    #     command = help_command

    # However, even if we aren't referring to the usage for the whole command, we can cross-check to ensure we're
    # using some correct arguments
    # if len(usage_command.positional) != len(help_command.positional):
    #     command.positional = usage_command.positional
    # confirmed = {pos.name for pos in usage_command.positional}
    # for positional in command.positional:
    #     if positional.name in confirmed:
    #         positional.usage_supported = True

    # Then, by default we filter out unsupported positionals, to remove false positives
    # if only_supported_positionals:
    #     command.positional = [pos for pos in command.positional if pos.usage_supported]

    return command


def best_cmd(
    cmd: typing.List[str],
    flags: typing.Iterable[str] = ([], ["-h"], ["--help"], ["--usage"]),
    run_kwargs: dict = {},
) -> Command:
    """
    Determine the best Command instance for a given command line tool, by trying many
    different help flags, such as --help and -h
    :param cmd: The command to analyse, e.g. ['wc'] or ['bwa', 'mem']
    :param flags: A list of help flags to try, e.g. ['--help', '-h']
    :param run_kwargs: kwargs to pass into subprocess.run, when we run the executable
    """
    # For each help flag, run the command and then try to parse it
    commands = []
    for flag in flags:
        help_cmd = cmd + flag
        try:
            final = execute_cmd(help_cmd, **run_kwargs)
            commands.append(parse_help(cmd, final))
        except (ParseBaseException, UnicodeDecodeError):
            # If parsing fails, this wasn't the right flag to use
            continue

    return max(commands, key=lambda com: len(com.named) + len(com.positional))


def explore_command(
    cmd: typing.List[str],
    flags: typing.Iterable[str] = ([], ["-h"], ["--help"], ["--usage"]),
    parent: typing.Optional[Command] = None,
    run_kwargs: dict = {},
) -> typing.Optional[Command]:
    """
    Given a command to start with, builds a model of this command and all its subcommands (if they exist)
    :param cmd: Command line executable and arguments to explore
    :param flags: List of flags to append to cmd in order to look for help commands, e.g. "--help"
    :param parent: A parent Command to add this command to as a subcommand, if this command actually exists
    :param run_kwargs: kwargs to pass into subprocess.run, when we run the executable
    """
    command = best_cmd(cmd, flags, run_kwargs=run_kwargs)

    if parent:
        # This isn't a subcommand if it has no flags
        if len(command.positional) + len(command.named) == 0:
            return None

        # This isn't a subcommand if it shares any positional with the parent command
        for pos_a, pos_b in zip(parent.positional, command.positional):
            if pos_a == pos_b:
                return None

        # This isn't a subcommand if it shares any flags with the parent command
        for flag_a, flag_b in zip(parent.named, command.named):
            if flag_a == flag_b:
                return None

    # Recursively call this function on positionals
    for positional in command.positional:
        subcommand = explore_command(
            cmd + [positional.name], flags=flags, parent=command, run_kwargs=run_kwargs
        )
        if subcommand is not None:
            command.subcommands.append(subcommand)

            # If we had any subcommands then we probably don't have any positionals, or at least don't care about them
            command.positional = []

    return command


def execute_cmd(help_cmd: typing.List[str], **kwargs) -> str:
    """
    Execute a command defined by a list of arguments, and return the result as a string
    """
    defaults = dict(stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=5)
    defaults.update(kwargs)
    try:
        proc = subprocess.run(help_cmd, **defaults)
        return (proc.stdout or proc.stderr).decode("utf_8")
    except subprocess.TimeoutExpired:
        return ""
